from __future__ import annotations

import os
import time
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from daft.execution.execution_step import PartitionTask


def get_tqdm(use_ray_tqdm: bool) -> Any:
    # Choose the appropriate tqdm module depending on whether we need to use Ray's tqdm
    if use_ray_tqdm:
        from ray.experimental.tqdm_ray import tqdm
    else:
        from tqdm.auto import tqdm as _tqdm

        try:
            import sys

            from IPython import get_ipython

            ipython = get_ipython()

            # write to sys.stdout if in jupyter notebook
            # source: https://github.com/tqdm/tqdm/blob/74722959a8626fd2057be03e14dcf899c25a3fd5/tqdm/autonotebook.py#L14
            if ipython is not None and "IPKernelApp" in ipython.config:

                class tqdm(_tqdm):  # type: ignore
                    def __init__(self, *args: Any, **kwargs: Any) -> None:
                        kwargs = kwargs.copy()
                        if "file" not in kwargs:
                            kwargs["file"] = sys.stdout  # avoid the red block in IPython

                        super().__init__(*args, **kwargs)
            else:
                tqdm = _tqdm
        except ImportError:
            tqdm = _tqdm

    return tqdm


class ProgressBar:
    def __init__(self, use_ray_tqdm: bool, show_tasks_bar: bool = False, disable: bool = False) -> None:
        self.show_tasks_bar = show_tasks_bar
        self._maxinterval = 5.0

        self.use_ray_tqdm = use_ray_tqdm
        self.tqdm_mod = get_tqdm(use_ray_tqdm)

        self.pbars: dict[int, Any] = dict()
        self.disable = (
            disable
            or not bool(int(os.environ.get("RAY_TQDM", "1")))
            or not bool(int(os.environ.get("DAFT_PROGRESS_BAR", "1")))
        )

    def _make_new_bar(self, stage_id: int, name: str) -> None:
        if self.use_ray_tqdm:
            self.pbars[stage_id] = self.tqdm_mod(total=1, desc=name, position=len(self.pbars))
        else:
            self.pbars[stage_id] = self.tqdm_mod(
                total=1,
                desc=name,
                position=len(self.pbars),
                leave=False,
                mininterval=1.0,
                maxinterval=self._maxinterval,
            )

    def mark_task_start(self, step: PartitionTask[Any]) -> None:
        if self.disable:
            return
        if self.show_tasks_bar:
            if len(self.pbars) == 0:
                self._make_new_bar(-1, "Tasks")
            else:
                task_pbar = self.pbars[-1]
                task_pbar.total += 1

        stage_id = step.stage_id

        if stage_id not in self.pbars:
            name = step.name()
            self._make_new_bar(stage_id, name)
        else:
            pb = self.pbars[stage_id]
            pb.total += 1
            if hasattr(pb, "last_print_t"):
                dt = time.time() - pb.last_print_t
                if dt >= self._maxinterval:
                    pb.refresh()

    def mark_task_done(self, step: PartitionTask[Any]) -> None:
        if self.disable:
            return
        stage_id = step.stage_id
        self.pbars[stage_id].update(1)
        if self.show_tasks_bar:
            self.pbars[-1].update(1)

    def close(self) -> None:
        for p in self.pbars.values():
            if not self.use_ray_tqdm:
                p.clear()
            p.close()
            del p


# Progress Bar for local execution, should only be used in the native executor
class SwordfishProgressBar:
    def __init__(self) -> None:
        self._maxinterval = 5.0
        self.tqdm_mod = get_tqdm(False)
        self.pbars: dict[int, Any] = dict()
        self.bar_configs: dict[int, str] = dict()
        self.next_id = 0

    def make_new_bar(self, bar_format: str) -> int:
        pbar_id = self.next_id
        self.next_id += 1
        self.bar_configs[pbar_id] = bar_format
        return pbar_id

    def update_bar(self, pbar_id: int, message: str) -> None:
        if pbar_id not in self.pbars:
            if pbar_id not in self.bar_configs:
                raise ValueError(f"No bar configuration found for id {pbar_id}")
            bar_format = self.bar_configs[pbar_id]
            self.pbars[pbar_id] = self.tqdm_mod(
                bar_format=bar_format,
                position=pbar_id,
                leave=False,
                mininterval=1.0,
                maxinterval=self._maxinterval,
            )
            del self.bar_configs[pbar_id]
        self.pbars[pbar_id].set_description_str(message)

    def close_bar(self, pbar_id: int) -> None:
        if pbar_id in self.pbars:
            self.pbars[pbar_id].close()
            del self.pbars[pbar_id]

    def close(self) -> None:
        for p in self.pbars.values():
            p.close()
            del p



from __future__ import annotations

import pytest

from daft.daft import io_glob

from .conftest import minio_create_bucket


def compare_s3_result(daft_ls_result: list, s3fs_result: list):
    daft_files = [(f["path"], f["type"].lower()) for f in daft_ls_result]
    s3fs_files = [(f"s3://{f['name']}", f["type"]) for f in s3fs_result]

    # io_glob does not return directories
    s3fs_files = [(p, t) for p, t in s3fs_files if t == "file"]

    assert sorted(daft_files) == sorted(s3fs_files)


def s3fs_recursive_list(fs, path) -> list:
    all_results = []
    curr_level_result = fs.ls(path, detail=True)
    for item in curr_level_result:
        if item["type"] == "directory":
            new_path = f's3://{item["name"]}'
            all_results.extend(s3fs_recursive_list(fs, new_path))
            item["name"] += "/"
            all_results.append(item)
        else:
            all_results.append(item)
    return all_results


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path_expect_pair",
    [
        # Exact filepath:
        ("s3://bucket/a.match", [{"type": "File", "path": "s3://bucket/a.match", "size": 0}]),
        # Exact filepath but with s3a:
        ("s3a://bucket/a.match", [{"type": "File", "path": "s3a://bucket/a.match", "size": 0}]),
        ###
        # `**`: recursive wildcard
        ###
        # All files with **
        (
            "s3://bucket/**",
            [
                {"type": "File", "path": "s3://bucket/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/b.nomatch", "size": 0},
                {"type": "File", "path": "s3://bucket/c.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/b.nomatch", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/c.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/b.nomatch", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/c.match", "size": 0},
            ],
        ),
        # All files with s3a and **
        (
            "s3a://bucket/**",
            [
                {"type": "File", "path": "s3a://bucket/a.match", "size": 0},
                {"type": "File", "path": "s3a://bucket/b.nomatch", "size": 0},
                {"type": "File", "path": "s3a://bucket/c.match", "size": 0},
                {"type": "File", "path": "s3a://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3a://bucket/nested1/b.nomatch", "size": 0},
                {"type": "File", "path": "s3a://bucket/nested1/c.match", "size": 0},
                {"type": "File", "path": "s3a://bucket/nested2/a.match", "size": 0},
                {"type": "File", "path": "s3a://bucket/nested2/b.nomatch", "size": 0},
                {"type": "File", "path": "s3a://bucket/nested2/c.match", "size": 0},
            ],
        ),
        # Exact filepath after **
        (
            "s3://bucket/**/a.match",
            [
                {"type": "File", "path": "s3://bucket/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/a.match", "size": 0},
            ],
        ),
        # Wildcard filepath after **
        (
            "s3://bucket/**/nested1/*.match",
            [
                {"type": "File", "path": "s3://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/c.match", "size": 0},
            ],
        ),
        # Wildcard folder before **
        (
            "s3://bucket/*/**/*.match",
            [
                {"type": "File", "path": "s3://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/c.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/c.match", "size": 0},
            ],
        ),
        ###
        # `*`: wildcard
        ###
        # Wildcard file
        (
            "s3://bucket/*.match",
            [
                {"type": "File", "path": "s3://bucket/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/c.match", "size": 0},
            ],
        ),
        # Wildcard file
        (
            "s3://bucket/*",
            [
                {"type": "File", "path": "s3://bucket/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/b.nomatch", "size": 0},
                {"type": "File", "path": "s3://bucket/c.match", "size": 0},
            ],
        ),
        # Nested wildcard file
        (
            "s3://bucket/nested1/*.match",
            [
                {"type": "File", "path": "s3://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/c.match", "size": 0},
            ],
        ),
        # Wildcard folder + wildcard file
        (
            "s3://bucket/*1/*.match",
            [
                {"type": "File", "path": "s3://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/c.match", "size": 0},
            ],
        ),
        # Wildcard folder + exact file
        (
            "s3://bucket/*/a.match",
            [
                {"type": "File", "path": "s3://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/a.match", "size": 0},
            ],
        ),
        ###
        # Missing paths
        ###
        # Exact filepath missing:
        ("s3://bucket/MISSING", FileNotFoundError),
        # Exact filepath missing before wildcard:
        ("s3://bucket/MISSING/*", FileNotFoundError),
        # Exact filepath missing after wildcard:
        ("s3://bucket/*/MISSING", []),
        # Wildcard file no match:
        ("s3://bucket/*.MISSING", []),
        # Wildcard folder no match:
        ("s3://bucket/*NOMATCH/*.match", []),
        ###
        # Directories: glob ignores directories and never returns them
        ###
        # Exact directory: fall back to ls behavior but ignore Directories
        (
            "s3://bucket",
            [
                {"type": "File", "path": "s3://bucket/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/b.nomatch", "size": 0},
                {"type": "File", "path": "s3://bucket/c.match", "size": 0},
            ],
        ),
        # Wildcard folder: we don't select directories with wildcards because we think it's a File
        ("s3://bucket/nested*", []),
        # Wildcard folder: Directories can be selected with a trailing /
        (
            "s3://bucket/nested*/",
            [
                {"type": "File", "path": "s3://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/b.nomatch", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/c.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/b.nomatch", "size": 0},
                {"type": "File", "path": "s3://bucket/nested2/c.match", "size": 0},
            ],
        ),
        # Exact folder after **: we don't return directories
        ("s3://bucket/**/nested1", []),
        # Exact folder after **: return results if user specifies a trailing /
        (
            "s3://bucket/**/nested1/",
            [
                {"type": "File", "path": "s3://bucket/nested1/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/b.nomatch", "size": 0},
                {"type": "File", "path": "s3://bucket/nested1/c.match", "size": 0},
            ],
        ),
    ],
)
@pytest.mark.parametrize("fanout_limit", [None, 1])
def test_directory_globbing_fragment_wildcard(minio_io_config, path_expect_pair, fanout_limit):
    globpath, expect = path_expect_pair
    with minio_create_bucket(minio_io_config, bucket_name="bucket") as fs:
        files = [
            "a.match",
            "b.nomatch",
            "c.match",
            "nested1/a.match",
            "nested1/b.nomatch",
            "nested1/c.match",
            "nested2/a.match",
            "nested2/b.nomatch",
            "nested2/c.match",
        ]
        for name in files:
            fs.touch(f"bucket/{name}")

        if type(expect) is type and issubclass(expect, BaseException):
            with pytest.raises(expect):
                io_glob(globpath, io_config=minio_io_config, fanout_limit=fanout_limit)
        else:
            daft_ls_result = io_glob(globpath, io_config=minio_io_config, fanout_limit=fanout_limit)
            assert sorted(daft_ls_result, key=lambda d: d["path"]) == sorted(expect, key=lambda d: d["path"])


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path_expect_pair",
    [
        # A "\*" is escaped to a literal *
        (r"s3://bucket/\*.match", [{"type": "File", "path": "s3://bucket/*.match", "size": 0}]),
        # A "\\" is escaped to just a \
        (r"s3://bucket/\\.match", [{"type": "File", "path": r"s3://bucket/\.match", "size": 0}]),
        # Ignore \ followed by non-special character
        (r"s3://bucket/\a.match", [{"type": "File", "path": "s3://bucket/a.match", "size": 0}]),
    ],
)
def test_directory_globbing_escape_characters(minio_io_config, path_expect_pair):
    globpath, expect = path_expect_pair
    with minio_create_bucket(minio_io_config, bucket_name="bucket") as fs:
        files = ["a.match", "*.match", r"\.match"]
        for name in files:
            fs.touch(f"bucket/{name}")
        daft_ls_result = io_glob(globpath, io_config=minio_io_config)
        assert sorted(daft_ls_result, key=lambda d: d["path"]) == sorted(expect, key=lambda d: d["path"])


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path_expect_pair",
    [
        # Test [] square brackets for matching specified single characters
        (
            "s3://bucket/[ab].match",
            [
                {"type": "File", "path": "s3://bucket/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/b.match", "size": 0},
            ],
        ),
        # Test ? for matching any single characters
        (
            "s3://bucket/?.match",
            [
                {"type": "File", "path": "s3://bucket/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/b.match", "size": 0},
                {"type": "File", "path": "s3://bucket/c.match", "size": 0},
                {"type": "File", "path": "s3://bucket/d.match", "size": 0},
            ],
        ),
        # Test {} for matching arbitrary globs
        (
            "s3://bucket/{a,[bc]}.match",
            [
                {"type": "File", "path": "s3://bucket/a.match", "size": 0},
                {"type": "File", "path": "s3://bucket/b.match", "size": 0},
                {"type": "File", "path": "s3://bucket/c.match", "size": 0},
            ],
        ),
    ],
)
def test_directory_globbing_special_characters(minio_io_config, path_expect_pair):
    globpath, expect = path_expect_pair
    with minio_create_bucket(minio_io_config, bucket_name="bucket") as fs:
        files = ["a.match", "b.match", "c.match", "d.match"]
        for name in files:
            fs.touch(f"bucket/{name}")
        daft_ls_result = io_glob(globpath, io_config=minio_io_config)
        assert sorted(daft_ls_result, key=lambda d: d["path"]) == sorted(expect, key=lambda d: d["path"])


@pytest.mark.integration()
def test_directory_globbing_common_prefix_cornercase(minio_io_config):
    with minio_create_bucket(minio_io_config, bucket_name="bucket") as fs:
        files = [
            "1/a/file.txt",
            "1/b/file.txt",
            # share a prefix with `1` which may cause issues if we drop the trailing / in a prefix list
            "11/a/file.txt",
            "11/b/file.txt",
        ]
        for name in files:
            fs.touch(f"bucket/{name}")

        # Force a prefix listing on the second level when the fanout becomes more than 2
        daft_ls_result = io_glob("s3://bucket/**", io_config=minio_io_config, fanout_limit=2)
        assert sorted(daft_ls_result, key=lambda d: d["path"]) == sorted(
            [
                {"type": "File", "path": "s3://bucket/1/a/file.txt", "size": 0},
                {"type": "File", "path": "s3://bucket/1/b/file.txt", "size": 0},
                {"type": "File", "path": "s3://bucket/11/a/file.txt", "size": 0},
                {"type": "File", "path": "s3://bucket/11/b/file.txt", "size": 0},
            ],
            key=lambda d: d["path"],
        )


@pytest.mark.integration()
def test_flat_directory_listing(minio_io_config):
    bucket_name = "bucket"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        files = ["a", "b", "c"]
        for name in files:
            fs.touch(f"{bucket_name}/{name}")
        daft_ls_result = io_glob(f"s3://{bucket_name}", io_config=minio_io_config)
        s3fs_result = fs.ls(f"s3://{bucket_name}", detail=True)
        compare_s3_result(daft_ls_result, s3fs_result)


@pytest.mark.integration()
def test_recursive_directory_listing(minio_io_config):
    bucket_name = "bucket"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        files = ["a", "b/bb", "c/cc/ccc"]
        for name in files:
            fs.write_bytes(f"s3://{bucket_name}/{name}", b"")
        daft_ls_result = io_glob(f"s3://{bucket_name}/**", io_config=minio_io_config)
        fs.invalidate_cache()
        s3fs_result = s3fs_recursive_list(fs, path=f"s3://{bucket_name}")
        compare_s3_result(daft_ls_result, s3fs_result)


@pytest.mark.integration()
def test_single_file_directory_listing(minio_io_config):
    bucket_name = "bucket"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        files = ["a", "b/bb", "c/cc/ccc"]
        for name in files:
            fs.write_bytes(f"s3://{bucket_name}/{name}", b"")
        daft_ls_result = io_glob(f"s3://{bucket_name}/c/cc/ccc", io_config=minio_io_config)
        fs.invalidate_cache()
        s3fs_result = s3fs_recursive_list(fs, path=f"s3://{bucket_name}/c/cc/ccc")
        assert len(daft_ls_result) == 1
        compare_s3_result(daft_ls_result, s3fs_result)


@pytest.mark.integration()
def test_single_file_directory_listing_trailing(minio_io_config):
    bucket_name = "bucket"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        files = ["a", "b/bb", "c/cc/ccc"]
        for name in files:
            fs.write_bytes(f"s3://{bucket_name}/{name}", b"")
        daft_ls_result = io_glob(f"s3://{bucket_name}/c/cc///", io_config=minio_io_config)
        fs.invalidate_cache()
        s3fs_result = s3fs_recursive_list(fs, path=f"s3://{bucket_name}/c/cc///")
        assert len(daft_ls_result) == 1
        compare_s3_result(daft_ls_result, s3fs_result)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "recursive",
    [False, True],
)
def test_missing_file_path(minio_io_config, recursive):
    bucket_name = "bucket"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        files = ["a", "b/bb", "c/cc/ccc"]
        path = f"s3://{bucket_name}/c/cc/ddd/**" if recursive else f"s3://{bucket_name}/c/cc/ddd"
        for name in files:
            fs.write_bytes(f"s3://{bucket_name}/{name}", b"")
        with pytest.raises(FileNotFoundError, match=f"s3://{bucket_name}/c/cc/ddd"):
            io_glob(path, io_config=minio_io_config)


@pytest.mark.integration()
@pytest.mark.parametrize("limit", [1, 2])
def test_limit(minio_io_config, limit):
    bucket_name = "bucket"
    with minio_create_bucket(minio_io_config, bucket_name=bucket_name) as fs:
        files = ["a", "b/bb", "c/cc/ccc"]
        for name in files:
            fs.write_bytes(f"s3://{bucket_name}/{name}", b"")
        daft_ls_result = io_glob(f"s3://{bucket_name}/**", io_config=minio_io_config, limit=limit)
        assert len(daft_ls_result) == limit



"""WARNING! These APIs are internal; please use Catalog.from_iceberg() and Table.from_iceberg()."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

from pyiceberg.catalog import Catalog as InnerCatalog
from pyiceberg.catalog import load_catalog
from pyiceberg.exceptions import NoSuchNamespaceError, NoSuchTableError
from pyiceberg.table import Table as InnerTable

from daft.catalog import Catalog, Identifier, NotFoundError, Properties, Schema, Table
from daft.io._iceberg import read_iceberg

if TYPE_CHECKING:
    from daft.dataframe import DataFrame


class IcebergCatalog(Catalog):
    _inner: InnerCatalog

    def __init__(self, pyiceberg_catalog: InnerCatalog):
        """DEPRECATED: Please use `Catalog.from_iceberg`; version 0.5.0!"""
        warnings.warn(
            "This is deprecated and will be removed in daft >= 0.5.0, please use `Catalog.from_iceberg` instead.",
            category=DeprecationWarning,
        )
        self._inner = pyiceberg_catalog

    @staticmethod
    def _from_obj(obj: object) -> IcebergCatalog:
        """Returns an IcebergCatalog instance if the given object can be adapted so."""
        if isinstance(obj, InnerCatalog):
            c = IcebergCatalog.__new__(IcebergCatalog)
            c._inner = obj
            return c
        raise ValueError(f"Unsupported iceberg catalog type: {type(obj)}")

    @staticmethod
    def _load_catalog(name: str, **options: str | None) -> IcebergCatalog:
        c = IcebergCatalog.__new__(IcebergCatalog)
        c._inner = load_catalog(name, **options)
        return c

    @property
    def name(self) -> str:
        return self._inner.name

    ###
    # create_*
    ###

    def _create_namespace(self, identifier: Identifier) -> None:
        ident = _to_pyiceberg_ident(identifier)
        self._inner.create_namespace(ident)

    def _create_table(
        self,
        identifier: Identifier,
        schema: Schema,
        properties: Properties | None = None,
    ) -> Table:
        i = _to_pyiceberg_ident(identifier)
        t = IcebergTable.__new__(IcebergTable)
        t._inner = self._inner.create_table(i, schema=schema.to_pyarrow_schema())
        return t

    ###
    # drop_*
    ###

    def _drop_namespace(self, identifier: Identifier) -> None:
        ident = _to_pyiceberg_ident(identifier)
        self._inner.drop_namespace(ident)

    def _drop_table(self, identifier: Identifier) -> None:
        ident = _to_pyiceberg_ident(identifier)
        self._inner.drop_table(ident)

    ###
    # has_*
    ###

    def _has_namespace(self, identifier: Identifier) -> bool:
        ident = _to_pyiceberg_ident(identifier)
        try:
            _ = self._inner.list_namespaces(ident)
            return True
        except NoSuchNamespaceError:
            return False

    def _has_table(self, identifier: Identifier) -> bool:
        ident = _to_pyiceberg_ident(identifier)
        try:
            # using load_table instead of table_exists because table_exists does not work with an instance of the `tabulario/iceberg-rest` Docker image
            self._inner.load_table(ident)
            return True
        except NoSuchTableError:
            return False

    ###
    # get_*
    ###

    def _get_table(self, identifier: Identifier) -> IcebergTable:
        ident = _to_pyiceberg_ident(identifier)
        try:
            return IcebergTable._from_obj(self._inner.load_table(ident))
        except NoSuchTableError as ex:
            # convert to not found because we want to (sometimes) ignore it internally
            raise NotFoundError() from ex
        except Exception as ex:
            # wrap original exceptions
            raise Exception("pyiceberg raised an exception while calling get_table") from ex

    ###
    # list_*
    ###

    def _list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        prefix = () if pattern is None else _to_pyiceberg_ident(pattern)
        return [Identifier(*tup) for tup in self._inner.list_namespaces(prefix)]

    def _list_tables(self, pattern: str | None = None) -> list[Identifier]:
        if pattern is None:
            tables = []
            for ns in self.list_namespaces():
                tables.extend(self._inner.list_tables(str(ns)))
        else:
            tables = self._inner.list_tables(pattern)
        return [Identifier(*tup) for tup in tables]


class IcebergTable(Table):
    _inner: InnerTable

    _read_options = {"snapshot_id"}
    _write_options: set[str] = set()

    def __init__(self, inner: InnerTable):
        """DEPRECATED: Please use `Table.from_iceberg`; version 0.5.0!"""
        warnings.warn(
            "This is deprecated and will be removed in daft >= 0.5.0, please prefer using `Table.from_iceberg` instead; version 0.5.0!",
            category=DeprecationWarning,
        )
        self._inner = inner

    @property
    def name(self) -> str:
        return self._inner.name()[-1]

    def schema(self) -> Schema:
        return self.read().schema()

    @staticmethod
    def _from_obj(obj: object) -> IcebergTable:
        """Returns an IcebergTable if the given object can be adapted so."""
        if isinstance(obj, InnerTable):
            t = IcebergTable.__new__(IcebergTable)
            t._inner = obj
            return t
        raise ValueError(f"Unsupported iceberg table type: {type(obj)}")

    def read(self, **options: Any | None) -> DataFrame:
        Table._validate_options("Iceberg read", options, IcebergTable._read_options)
        return read_iceberg(self._inner, snapshot_id=options.get("snapshot_id"))

    def append(self, df: DataFrame, **options: Any) -> None:
        self._validate_options("Iceberg write", options, IcebergTable._write_options)

        df.write_iceberg(self._inner, mode="append")

    def overwrite(self, df: DataFrame, **options: Any) -> None:
        self._validate_options("Iceberg write", options, IcebergTable._write_options)

        df.write_iceberg(self._inner, mode="overwrite")


def _to_pyiceberg_ident(ident: Identifier | str) -> tuple[str, ...] | str:
    return tuple(ident) if isinstance(ident, Identifier) else ident



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from typing import Optional

from daft import context
from daft.api_annotations import PublicAPI
from daft.daft import IOConfig, ScanOperatorHandle, StorageConfig
from daft.dataframe import DataFrame
from daft.logical.builder import LogicalPlanBuilder


@PublicAPI
def read_hudi(
    table_uri: str,
    io_config: Optional[IOConfig] = None,
) -> DataFrame:
    """Create a DataFrame from a Hudi table.

    Args:
        table_uri: URI to the Hudi table.
        io_config: A custom IOConfig to use when accessing Hudi table object storage data. Defaults to None.

    Returns:
        DataFrame: A DataFrame with the schema converted from the specified Hudi table.

    Examples:
        >>> df = daft.read_hudi("some-table-uri")
        >>> df = df.where(df["foo"] > 5)
        >>> df.show()
    """
    from daft.hudi.hudi_scan import HudiScanOperator

    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config

    multithreaded_io = context.get_context().get_or_create_runner().name != "ray"
    storage_config = StorageConfig(multithreaded_io, io_config)

    hudi_operator = HudiScanOperator(table_uri, storage_config=storage_config)

    handle = ScanOperatorHandle.from_python_scan_operator(hudi_operator)
    builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)
    return DataFrame(builder)



from __future__ import annotations

import pytest


def test_print_schema_range(spark_session, capsys) -> None:
    df = spark_session.range(10)
    df.printSchema()

    captured = capsys.readouterr()
    expected = "root\n |-- id: long (nullable = true)\n\n"
    assert captured.out == expected


def test_print_schema_simple_df(spark_session, capsys) -> None:
    data = [(1,), (2,), (3,)]
    df = spark_session.createDataFrame(data, ["value"])
    df.printSchema()

    captured = capsys.readouterr()
    expected = "root\n |-- value: long (nullable = true)\n\n"
    assert captured.out == expected


def test_print_schema_multiple_columns(spark_session, capsys) -> None:
    data = [(1, "a", True), (2, "b", False)]
    df = spark_session.createDataFrame(data, ["id", "name", "flag"])
    df.printSchema()

    captured = capsys.readouterr()
    expected = (
        "root\n"
        " |-- id: long (nullable = true)\n"
        " |-- name: string (nullable = true)\n"
        " |-- flag: boolean (nullable = true)\n\n"
    )
    assert captured.out == expected


def test_print_schema_floating_point(spark_session, capsys) -> None:
    data = [(1.23,), (4.56,)]
    df = spark_session.createDataFrame(data, ["amount"])
    df.printSchema()

    captured = capsys.readouterr()
    expected = "root\n |-- amount: double (nullable = true)\n\n"
    assert captured.out == expected


def test_print_schema_with_nulls(spark_session, capsys) -> None:
    data = [(1, None), (None, "test")]
    df = spark_session.createDataFrame(data, ["id", "value"])
    df.printSchema()

    captured = capsys.readouterr()
    expected = "root\n |-- id: long (nullable = true)\n |-- value: string (nullable = true)\n\n"
    assert captured.out == expected


@pytest.mark.skip(
    reason="Skipping due to https://github.com/Eventual-Inc/Daft/issues/3605 - conversion doesn't work properly for nested structs"
)
def test_print_schema_nested(spark_session) -> None:
    nested_data = [(1, {"name": "John", "age": 30}), (2, {"name": "Jane", "age": 25})]

    # Create DataFrame with nested structures
    df = spark_session.createDataFrame(nested_data, ["id", "info"])

    # Print schema
    print("DataFrame Schema:")
    df.printSchema()



from __future__ import annotations

import functools
import inspect
import sys
from typing import Any, Callable, ForwardRef, Literal, TypeVar, Union, get_args, get_origin

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

from daft.analytics import time_df_method, time_func

T = TypeVar("T")
P = ParamSpec("P")


def DataframePublicAPI(func: Callable[P, T]) -> Callable[P, T]:
    """A decorator to mark a function as part of the Daft DataFrame's public API."""

    @functools.wraps(func)
    def _wrap(*args: P.args, **kwargs: P.kwargs) -> T:
        type_check_function(func, *args, **kwargs)
        timed_method = time_df_method(func)
        return timed_method(*args, **kwargs)

    return _wrap


def PublicAPI(func: Callable[P, T]) -> Callable[P, T]:
    """A decorator to mark a function as part of the Daft public API."""

    @functools.wraps(func)
    def _wrap(*args: P.args, **kwargs: P.kwargs) -> T:
        __tracebackhide__ = True
        type_check_function(func, *args, **kwargs)
        timed_func = time_func(func)
        return timed_func(*args, **kwargs)

    return _wrap


class APITypeError(TypeError):
    pass


def type_check_function(func: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
    signature = inspect.signature(func)
    arguments = signature.bind(*args, **kwargs).arguments
    type_hints = func.__annotations__

    def isinstance_helper(value: Any, T: Any) -> bool:
        """Like builtins.isinstance, but also accepts typing.* types."""
        if T is Any:
            return True

        # T is an unresolved annotation.
        # We cannot typecheck these, so just treat them as Any.
        if isinstance(T, (str, ForwardRef)):
            return True

        origin_T = get_origin(T)

        # T is a builtin primitive type, like `int`
        if origin_T is None:
            return isinstance(value, T)

        # T is a generic type, like `typing.List` or builtin container like `list`
        if isinstance(origin_T, type):
            return isinstance(value, origin_T)

        # T is a `typing.Union`
        if origin_T is Union:
            union_types = get_args(T)
            return any(isinstance_helper(value, union_type) for union_type in union_types)

        # T is a `typing.Literal`
        if origin_T is Literal:
            literal_values = get_args(T)
            return value in literal_values

        raise NotImplementedError(
            f"Unexpected error: Type checking is not implemented for type {T}. Sorry! Please file an issue."
        )

    for name, value in arguments.items():
        if name not in type_hints:
            continue

        param_kind = signature.parameters[name].kind
        param_type = type_hints[name]

        # Non-variadic arguments.
        if param_kind in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
        ):
            if not isinstance_helper(value, param_type):
                raise APITypeError(
                    f"{func.__qualname__} received wrong input type.\n"
                    f"Required:\n\t{name} = <{param_type}>\n"
                    f"Given:\n\t{name} = <{type(value).__name__}>"
                )

        elif param_kind == inspect.Parameter.VAR_POSITIONAL:
            for i, item in enumerate(value):
                if not isinstance_helper(item, param_type):
                    raise APITypeError(
                        f"{func.__qualname__} received wrong input type.\n"
                        f"Required:\n\t{name} = <{param_type}>, ...\n"
                        f"Given:\n\t{name} = <{type(item).__name__}> (in position {i})"
                    )

        elif param_kind == inspect.Parameter.VAR_KEYWORD:
            for key, item in value.items():
                if not isinstance_helper(item, param_type):
                    raise APITypeError(
                        f"{func.__qualname__} received wrong input type.\n"
                        f"Required:\n\t{key} = <{param_type}>\n"
                        f"Given:\n\t{key} = <{type(item).__name__}>"
                    )



from __future__ import annotations

import datetime

import pytest

import daft
from daft import col, interval
from daft.sql.sql import SQLCatalog


def test_nested():
    df = daft.from_pydict(
        {
            "A": [1, 2, 3, 4],
            "B": [1.5, 2.5, 3.5, 4.5],
            "C": [True, True, False, False],
            "D": [None, None, None, None],
        }
    )

    actual = daft.sql("SELECT (A + 1) AS try_this FROM df").collect()
    expected = df.select((daft.col("A") + 1).alias("try_this")).collect()

    assert actual.to_pydict() == expected.to_pydict()

    actual = daft.sql("SELECT *, (A + 1) AS try_this FROM df").collect()
    expected = df.with_column("try_this", df["A"] + 1).collect()

    assert actual.to_pydict() == expected.to_pydict()


def test_hash_exprs():
    df = daft.from_pydict(
        {
            "a": ["foo", "bar", "baz", "qux"],
            "ints": [1, 2, 3, 4],
            "floats": [1.5, 2.5, 3.5, 4.5],
        }
    )

    actual = (
        daft.sql("""
    SELECT
        hash(a) as hash_a,
        hash(a, 0) as hash_a_0,
        hash(a, seed:=0) as hash_a_seed_0,
        minhash(a, num_hashes:=10, ngram_size:= 100, seed:=10) as minhash_a,
        minhash(a, num_hashes:=10, ngram_size:= 100) as minhash_a_no_seed,
        minhash(a, num_hashes:=10, ngram_size:= 100, seed:=10, hash_function:='xxhash') as minhash_a_xxhash,
    FROM df
    """)
        .collect()
        .to_pydict()
    )

    expected = (
        df.select(
            col("a").hash().alias("hash_a"),
            col("a").hash(0).alias("hash_a_0"),
            col("a").hash(seed=0).alias("hash_a_seed_0"),
            col("a").minhash(num_hashes=10, ngram_size=100, seed=10).alias("minhash_a"),
            col("a").minhash(num_hashes=10, ngram_size=100).alias("minhash_a_no_seed"),
            col("a").minhash(num_hashes=10, ngram_size=100, seed=10, hash_function="xxhash").alias("minhash_a_xxhash"),
        )
        .collect()
        .to_pydict()
    )

    assert actual == expected

    with pytest.raises(Exception, match="Invalid arguments for minhash"):
        daft.sql("SELECT minhash() as hash_a FROM df").collect()

    with pytest.raises(Exception, match="num_hashes is required"):
        daft.sql("SELECT minhash(a) as hash_a FROM df").collect()


def test_count_star():
    df = daft.from_pydict(
        {
            "a": [1, 2, 3, 4],
        }
    )

    actual = daft.sql("SELECT COUNT(*) FROM df").collect()
    expected = df.agg(daft.col("*").count().alias("count")).collect()
    assert actual.to_pydict() == expected.to_pydict()


def test_between():
    df = daft.from_pydict(
        {
            "integers": [0, 1, 2, 3, 5],
        }
    )

    actual = daft.sql("SELECT * FROM df where integers between 1 and 4").collect().to_pydict()

    expected = df.filter(col("integers").between(1, 4)).collect().to_pydict()
    assert actual == expected


def test_is_in():
    df = daft.from_pydict({"idx": [1, 2, 3], "val": ["foo", "bar", "baz"]})
    expected = df.filter(col("val").is_in(["bar", "foo"])).collect().to_pydict()
    actual = daft.sql("select * from df where val in ('bar','foo')").collect().to_pydict()
    assert actual == expected

    # test negated too
    expected = df.filter(~col("val").is_in(["bar", "foo"])).collect().to_pydict()
    actual = daft.sql("select * from df where val not in ('bar','foo')").collect().to_pydict()
    assert actual == expected


def test_is_in_exprs():
    df = daft.from_pydict({"x": [1, 2, 3, 5, 9]})
    expected = {"x": [1, 2, 9]}
    catalog = SQLCatalog({"df": df})
    actual = daft.sql("select * from df where x in (0 + 1, 0 + 2, 0 + 9)", catalog).collect().to_pydict()

    assert actual == expected


def test_is_in_edge_cases():
    df = daft.from_pydict(
        {
            "nums": [1, 2, 3, None, 4, 5],
            "strs": ["a", "b", None, "c", "d", "e"],
        }
    )

    # Test with NULL values in the column
    actual = daft.sql("SELECT * FROM df WHERE strs IN ('a', 'b')").collect().to_pydict()
    expected = df.filter(col("strs").is_in(["a", "b"])).collect().to_pydict()
    assert actual == expected

    # Test with empty IN list
    with pytest.raises(Exception, match="Expected: an expression"):
        daft.sql("SELECT * FROM df WHERE nums IN ()").collect()

    # Test with numbers and NULL in IN list
    actual = daft.sql("SELECT * FROM df WHERE nums IN (1, NULL, 3)").collect().to_pydict()
    expected = df.filter(col("nums").is_in([1, None, 3])).collect().to_pydict()
    assert actual == expected

    # Test with single value
    actual = daft.sql("SELECT * FROM df WHERE nums IN (1)").collect().to_pydict()
    expected = df.filter(col("nums").is_in([1])).collect().to_pydict()

    # Test with mixed types in the IN list
    with pytest.raises(Exception, match="arguments to be of the same type"):
        daft.sql("SELECT * FROM df WHERE nums IN (1, '2', 3.0)").collect().to_pydict()


@pytest.mark.parametrize(
    "date_values, ts_values, expected_intervals",
    [
        (
            ["2022-01-01", "2020-02-29", "2029-05-15"],
            ["2022-01-01 10:00:00", "2020-02-29 23:59:59", "2029-05-15 12:34:56"],
            {
                "date_add_day": [
                    datetime.date(2022, 1, 2),
                    datetime.date(2020, 3, 1),
                    datetime.date(2029, 5, 16),
                ],
                "date_sub_month": [
                    datetime.date(2021, 12, 1),
                    datetime.date(2020, 1, 31),
                    datetime.date(2029, 4, 14),
                ],
                "ts_sub_year": [
                    datetime.datetime(2021, 1, 1, 10),
                    datetime.datetime(2019, 2, 28, 23, 59, 59),
                    datetime.datetime(2028, 5, 15, 12, 34, 56),
                ],
                "ts_add_hour": [
                    datetime.datetime(2022, 1, 1, 11, 0, 0),
                    datetime.datetime(2020, 3, 1, 0, 59, 59),
                    datetime.datetime(2029, 5, 15, 13, 34, 56),
                ],
                "ts_sub_minute": [
                    datetime.datetime(2022, 1, 1, 9, 57, 21),
                    datetime.datetime(2020, 2, 29, 23, 57, 20),
                    datetime.datetime(2029, 5, 15, 12, 32, 17),
                ],
                "ts_add_hour_mul_2": [
                    datetime.datetime(2022, 1, 1, 12, 0, 0),
                    datetime.datetime(2020, 3, 1, 1, 59, 59),
                    datetime.datetime(2029, 5, 15, 14, 34, 56),
                ],
                "ts_add_2_mul_hour": [
                    datetime.datetime(2022, 1, 1, 12, 0, 0),
                    datetime.datetime(2020, 3, 1, 1, 59, 59),
                    datetime.datetime(2029, 5, 15, 14, 34, 56),
                ],
                "ts_add_hour_mul_neg_2": [
                    datetime.datetime(2022, 1, 1, 8, 0),
                    datetime.datetime(2020, 2, 29, 21, 59, 59),
                    datetime.datetime(2029, 5, 15, 10, 34, 56),
                ],
                "ts_add_minute_mul_2": [
                    datetime.datetime(2022, 1, 1, 10, 2, 0),
                    datetime.datetime(2020, 3, 1, 0, 1, 59),
                    datetime.datetime(2029, 5, 15, 12, 36, 56),
                ],
                "ts_add_month_mul_2": [
                    datetime.datetime(2022, 3, 1, 10, 0, 0),
                    datetime.datetime(2020, 4, 29, 23, 59, 59),
                    datetime.datetime(2029, 7, 15, 12, 34, 56),
                ],
                "ts_add_year_mul_2": [
                    datetime.datetime(2024, 1, 1, 10, 0, 0),
                    datetime.datetime(2022, 3, 1, 23, 59, 59),
                    datetime.datetime(2031, 5, 15, 12, 34, 56),
                ],
            },
        ),
    ],
)
def test_interval_comparison(date_values, ts_values, expected_intervals):
    # Create DataFrame with date and timestamp columns
    df = daft.from_pydict({"date": date_values, "ts": ts_values}).select(
        col("date").cast(daft.DataType.date()), col("ts").str.to_datetime("%Y-%m-%d %H:%M:%S")
    )
    catalog = SQLCatalog({"test": df})

    expected_df = (
        df.select(
            (col("date") + interval(days=1)).alias("date_add_day"),
            (col("date") - interval(months=1)).alias("date_sub_month"),
            (col("ts") - interval(years=1, days=0)).alias("ts_sub_year"),
            (col("ts") + interval(hours=1)).alias("ts_add_hour"),
            (col("ts") - interval(minutes=1, seconds=99)).alias("ts_sub_minute"),
            (col("ts") + interval(hours=1) * 2).alias("ts_add_hour_mul_2"),
            (col("ts") + 2 * interval(hours=1)).alias("ts_add_2_mul_hour"),
            (col("ts") + interval(hours=1) * -2).alias("ts_add_hour_mul_neg_2"),
            (col("ts") + interval(minutes=1) * 2).alias("ts_add_minute_mul_2"),
            (col("ts") + interval(months=1) * 2).alias("ts_add_month_mul_2"),
            (col("ts") + interval(years=1, days=0) * 2).alias("ts_add_year_mul_2"),
        )
        .collect()
        .to_pydict()
    )

    actual_sql = (
        daft.sql(
            """
        SELECT
            date + INTERVAL '1' day AS date_add_day,
            date - INTERVAL '1 months' AS date_sub_month,
            ts - INTERVAL '1 year 0 days' AS ts_sub_year,
            ts + INTERVAL '1' hour AS ts_add_hour,
            ts - INTERVAL '1 minutes 99 second' AS ts_sub_minute,
            ts + INTERVAL '1' hour * 2 AS ts_add_hour_mul_2,
            ts + INTERVAL '1' hour * -2 AS ts_add_hour_mul_neg_2,
            ts + 2 * INTERVAL '1' hour AS ts_add_2_mul_hour,
            ts + INTERVAL '1' minute * 2 AS ts_add_minute_mul_2,
            ts + INTERVAL '1' month * 2 AS ts_add_month_mul_2,
            ts + INTERVAL '1' year * 2 AS ts_add_year_mul_2,
        FROM test
        """,
            catalog=catalog,
        )
        .collect()
        .to_pydict()
    )

    assert expected_df == actual_sql == expected_intervals


def test_coalesce():
    df = daft.from_pydict(
        {
            "a": [None, None, 3, None],
            "b": [None, 2, 4, None],
            "c": [None, None, 5, 6],
        }
    )

    expected = df.select(daft.coalesce(col("a"), col("b"), col("c")).alias("result")).to_pydict()

    catalog = SQLCatalog({"df": df})
    actual = daft.sql(
        """
    SELECT
        COALESCE(a, b, c) as result
    FROM df
    """,
        catalog=catalog,
    ).to_pydict()

    assert actual == expected


@pytest.mark.parametrize(
    "precision, value, expected",
    [
        (None, 3.14159, 3),
        (None, 3, 3),
        (1, 3.14159, 3.1),
        (2, 3.14159, 3.14),
    ],
)
def test_round(precision, value, expected):
    if precision is None:
        query = f"select round({value})"
    else:
        query = f"select round({value}, {precision})"
    actual = daft.sql(query).to_pydict()
    expected = {"literal": [expected]}

    assert actual == expected


# just another sanity check to make sure args & kwargs are working properly
@pytest.mark.parametrize(
    "query, should_work",
    [
        ("round(3.14159, decimals:=3)", True),
        ("round(input:=3.14159, decimals:=3)", True),
        ("round(input:=3.14159, 3)", False),
        ("round(3.111)", True),
        ("round(3.1111, 2)", True),
        ("round(decimals:=2, input:=3.14)", True),
        ("round(decimals:=2, 3.14)", False),
    ],
)
def test_round_arg_handling(query, should_work):
    query = f"select {query}"
    try:
        daft.sql(query).to_pydict()
        if not should_work:
            pytest.fail()
    except Exception:
        if should_work:
            pytest.fail()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import boto3
import botocore
import pytest
from moto import mock_aws

from daft.catalog import Catalog, Identifier, NotFoundError
from daft.catalog.__glue import (
    GlueCatalog,
    GlueCsvTable,
    GlueIcebergTable,
    GlueParquetTable,
    GlueTable,
    _convert_glue_schema,
    load_glue,
)
from daft.dataframe import DataFrame
from daft.logical.schema import DataType, Schema

if TYPE_CHECKING:
    from mypy_boto3_glue import GlueClient
    from mypy_boto3_glue.type_defs import ColumnOutputTypeDef as GlueColumnInfo
    from mypy_boto3_glue.type_defs import TableTypeDef as GlueTableInfo

    from daft.dataframe import DataFrame
else:
    GlueClient = Any
    GlueColumnInfo = Any
    GlueTableInfo = Any


@pytest.fixture
def mock_boto3_session():
    with mock_aws():
        yield boto3.Session(region_name="us-west-2")


@pytest.fixture
def mock_botocore_session():
    with mock_aws():
        sess = botocore.session.Session()
        sess.set_config_variable("region", "us-west-2")
        yield sess


@pytest.fixture
def glue_client():
    with mock_aws():
        yield boto3.client("glue", region_name="us-west-2")


@pytest.fixture
def glue_catalog(glue_client):
    gc = GlueCatalog.from_client("mock_glue_catalog", glue_client)
    gc._table_impls.append(GlueTestTable)  # !! REGISTER GLUE TEST TABLE !!
    return gc


###
# catalog constructors
###


def test_load_glue():
    catalog = load_glue(
        "mock_glue_catalog",
        region_name="us-west-2",
        api_version="2017-03-31",
        use_ssl=False,
        verify=True,
        endpoint_url="http://localhost",
        aws_access_key_id="test_access_key",
        aws_secret_access_key="test_secret_key",
        aws_session_token="test_session_token",
    )
    assert catalog.name == "mock_glue_catalog"


def test_catalog_from_client(glue_client):
    assert Catalog.from_glue("gc", client=glue_client)
    assert GlueCatalog.from_client("gc", glue_client)


def test_catalog_from_session(mock_boto3_session, mock_botocore_session):
    # Test both boto3 and botocore sessions.
    assert Catalog.from_glue("gc", session=mock_boto3_session)
    assert GlueCatalog.from_session("gc", session=mock_boto3_session)

    assert Catalog.from_glue("gc", session=mock_botocore_session)
    assert GlueCatalog.from_session("gc", session=mock_botocore_session)


def test_catalog_no_args():
    with pytest.raises(ValueError, match="Must provide either a client or session."):
        Catalog.from_glue("foo")


def test_glue_catalog_init():
    with pytest.raises(ValueError, match="not supported"):
        GlueCatalog()


###
# catalog methods
###


def test_create_namespace(glue_catalog, glue_client):
    # create a namespace with catalog
    glue_catalog.create_namespace("test_namespace")

    # verify with client
    res = glue_client.get_databases()
    assert len(res["DatabaseList"]) == 1
    assert res["DatabaseList"][0]["Name"] == "test_namespace"

    # err. already exists
    with pytest.raises(ValueError, match="already exists"):
        glue_catalog.create_namespace("test_namespace")


def test_create_namespace_if_not_exists(glue_catalog, glue_client):
    # create a namespace with catalog
    glue_catalog.create_namespace_if_not_exists("test_namespace")
    glue_catalog.create_namespace_if_not_exists("test_namespace")

    # verify we only have one with client
    res = glue_client.get_databases()
    assert len(res["DatabaseList"]) == 1


def test_drop_namespace(glue_catalog, glue_client):
    # create then drop the namespace
    glue_catalog.create_namespace("test_namespace")
    glue_catalog.drop_namespace("test_namespace")

    # verify was dropped with client
    res = glue_client.get_databases()
    assert len(res["DatabaseList"]) == 0

    # drop if doesn't exist should error
    with pytest.raises(NotFoundError, match="not found"):
        glue_catalog.drop_namespace("nonexistent_namespace")


def test_list_tables(glue_catalog, glue_client):
    # create some tables
    create_glue_database(glue_client, "test_database")
    create_glue_table(glue_client, "test_database", "table1")
    create_glue_table(glue_client, "test_database", "table2")

    # list tables with catalog
    tables = glue_catalog.list_tables("test_database")
    assert len(tables) == 2
    assert Identifier("test_database", "table1") in tables
    assert Identifier("test_database", "table2") in tables

    # test listing tables with pattern
    tables = glue_catalog.list_tables("test_database.table1")
    assert len(tables) == 1
    assert Identifier("test_database", "table1") in tables

    # test listing tables with no pattern
    with pytest.raises(ValueError, match="requires the pattern to contain a namespace"):
        glue_catalog.list_tables(None)


def test_list_namespaces(glue_catalog, glue_client):
    # create namespaces with the client
    create_glue_database(glue_client, "namespace1")
    create_glue_database(glue_client, "namespace2")

    # list namespaces with the catalog
    namespaces = glue_catalog.list_namespaces()
    assert len(namespaces) == 2
    assert Identifier("namespace1") in namespaces
    assert Identifier("namespace2") in namespaces

    # glue doesn't have any patterns/filters for get_databases
    with pytest.raises(ValueError):
        glue_catalog.list_namespaces("namespace1")


def test_get_table(glue_catalog, glue_client):
    # create a table with the client
    create_glue_database(glue_client, "test_database")
    create_glue_table(glue_client, "test_database", "test_table")

    # Get the table
    table = glue_catalog.get_table("test_database.test_table")
    assert table.name == "test_table"

    # Test getting a table that doesn't exist
    with pytest.raises(NotFoundError, match="not found"):
        glue_catalog.get_table("test_database.nonexistent_table")

    # Invalid identifiers
    with pytest.raises(ValueError, match="Expected identifier with form `<database_name>.<table_name>`"):
        glue_catalog.drop_table("a.b.c")
    with pytest.raises(ValueError, match="Expected identifier with form `<database_name>.<table_name>`"):
        glue_catalog.drop_table("a")


def test_get_table_no_impls(glue_catalog, glue_client):
    create_glue_database(glue_client, "test_database")
    create_glue_table(glue_client, "test_database", "test_table")

    # no impls!
    glue_catalog._table_impls = []

    # get_table should fail
    with pytest.raises(ValueError, match="No supported table implementation"):
        glue_catalog.get_table("test_database.test_table")


def test_drop_table(glue_catalog, glue_client):
    # create a table with the client
    create_glue_database(glue_client, "test_database")
    create_glue_table(glue_client, "test_database", "test_table")

    # drop the table with the catalog
    glue_catalog.drop_table("test_database.test_table")

    with pytest.raises(NotFoundError):
        glue_catalog.get_table("test_database.test_table")
    with pytest.raises(NotFoundError):
        glue_catalog.drop_table("test_database.nonexistent_table")

    with pytest.raises(ValueError, match="Expected identifier with form `<database_name>.<table_name>`"):
        glue_catalog.drop_table("a.b.c")
    with pytest.raises(ValueError, match="Expected identifier with form `<database_name>.<table_name>`"):
        glue_catalog.drop_table("a")


###
# helpers
###


def create_glue_database(client, database_name):
    client.create_database(
        DatabaseInput={
            "Name": database_name,
        }
    )


def create_glue_table(client, database_name, table_name, location=None):
    if location is None:
        location = f"s3://bucket/{table_name}/"
    client.create_table(
        DatabaseName=database_name,
        TableInput={
            "Name": table_name,
            "StorageDescriptor": {
                "Location": location,
            },
            "TableType": "EXTERNAL_TABLE",
            "Parameters": {
                # !! marker that this is a test table !!
                "pytest": "True"
            },
        },
    )


class GlueTestTable(GlueTable):
    """GlueTestTable shows how we register custom table implementations."""

    @classmethod
    def from_table_info(cls, catalog: GlueCatalog, table: GlueTableInfo) -> GlueTable:
        if bool(table["Parameters"].get("pytest")):
            return cls(catalog, table)
        raise ValueError("Expected Parameter pytest='True'")

    def read(self, **options) -> DataFrame:
        raise NotImplementedError

    def append(
        self,
        df: DataFrame,
        **options,
    ) -> None:
        raise NotImplementedError

    def overwrite(
        self,
        df: DataFrame,
        **options,
    ) -> None:
        raise NotImplementedError


###
# GlueCsvTable Testing
###


def test_glue_csv_table_init_not_supported():
    with pytest.raises(ValueError, match="not supported"):
        GlueCsvTable()


def test_glue_csv_table_from_table_info_1(glue_catalog):
    table_info = {
        "Name": "test_csv_table",
        "StorageDescriptor": {
            "Columns": [
                {"Name": "col1", "Type": "string"},
                {"Name": "col2", "Type": "integer"},
                {"Name": "col3", "Type": "double"},
            ],
            "Location": "s3://bucket/test_csv_table/",
        },
        "Parameters": {
            "classification": "CSV",
        },
    }
    csv_table = GlueCsvTable.from_table_info(glue_catalog, table_info)
    assert csv_table._path == "s3://bucket/test_csv_table/"
    assert len(csv_table._schema) == 3
    assert csv_table._schema == Schema.from_pydict(
        {
            "col1": DataType.string(),
            "col2": DataType.int32(),
            "col3": DataType.float64(),
        }
    )
    assert csv_table._has_headers is False
    assert csv_table._delimiter == ","


def test_glue_csv_table_from_table_info_2(glue_catalog):
    table_info = {
        "Name": "test_csv_table",
        "StorageDescriptor": {
            "Columns": [
                {"Name": "col1", "Type": "string"},
            ],
            "Location": "s3://bucket/test_csv_table/",
        },
        "Parameters": {
            "classification": "CSV",
            "skip.header.line.count": "0",
            "delimiter": "\t",
        },
    }
    csv_table = GlueCsvTable.from_table_info(glue_catalog, table_info)
    assert csv_table._has_headers is False
    assert csv_table._delimiter == "\t"


def test_glue_csv_table_from_table_info_3(glue_catalog):
    table_info = {
        "Name": "test_csv_table",
        "Parameters": {},  # !! ERR !! no classification
    }
    with pytest.raises(
        ValueError,
        match="GlueTableInfo is missing the required parameter 'classification'",
    ):
        GlueCsvTable.from_table_info(glue_catalog, table_info)


def test_glue_csv_table_from_table_info_4(glue_catalog):
    table_info = {
        "Name": "test_csv_table",
        "Parameters": {
            "classification": "parquet",
        },
    }
    with pytest.raises(ValueError, match="GlueTableInfo had classification parquet, but expected 'CSV'"):
        GlueCsvTable.from_table_info(glue_catalog, table_info)


###
# GlueParquetTable
###


def test_glue_parquet_table_init_not_supported():
    with pytest.raises(ValueError, match="not supported"):
        GlueParquetTable()


def test_glue_parquet_table_from_table_info_1(glue_catalog):
    table_info = {
        "Name": "test_parquet_table",
        "StorageDescriptor": {
            "Columns": [
                {"Name": "col1", "Type": "string"},
            ],
            "Location": "s3://bucket/test_parquet_table/",
        },
        "Parameters": {
            "classification": "parquet",
        },
    }
    parquet_table = GlueParquetTable.from_table_info(glue_catalog, table_info)
    assert parquet_table.name == "test_parquet_table"


def test_glue_parquet_table_from_table_info_2(glue_catalog):
    table_info = {
        "Name": "test_parquet_table",
        "Parameters": {},  # !! ERR !! no classification
    }
    with pytest.raises(
        ValueError,
        match="GlueTableInfo is missing the required parameter 'classification'",
    ):
        GlueParquetTable.from_table_info(glue_catalog, table_info)


def test_glue_parquet_table_from_table_info_3(glue_catalog):
    table_info = {
        "Name": "test_parquet_table",
        "Parameters": {
            "classification": "CSV",
        },
    }
    with pytest.raises(ValueError, match="GlueTableInfo had classification CSV, but expected 'parquet'"):
        GlueParquetTable.from_table_info(glue_catalog, table_info)


###
# GlueIcebergTable
###


def test_glue_iceberg_table_init_not_supported():
    with pytest.raises(ValueError, match="not supported"):
        GlueIcebergTable()


def test_glue_iceberg_table_from_table_info_1(glue_catalog):
    table_info = {
        "DatabaseName": "test_iceberg_database",
        "Name": "test_iceberg_table",
        "StorageDescriptor": {},
        "Parameters": {
            "table_type": "iceberg",
            "metadata_location": "s3://bucket/test_iceberg_table/metadata.json",
        },
    }
    with pytest.raises(Exception):
        # Fails in pyarrow as OSError, which is ok for testing because it means we made it that far.
        # i.e. the daft logic is correctly picking up the metadata and passing to pyiceberg/pyarrow.
        GlueIcebergTable.from_table_info(glue_catalog, table_info)


def test_glue_iceberg_table_from_table_info_2(glue_catalog):
    table_info = {
        "Name": "test_iceberg_table",
        "Parameters": {},  # !! ERR !! no table_type
    }
    with pytest.raises(ValueError, match="missing the required 'table_type' parameter"):
        GlueIcebergTable.from_table_info(glue_catalog, table_info)


def test_glue_iceberg_table_from_table_info_3(glue_catalog):
    table_info = {
        "Name": "test_iceberg_table",
        "Parameters": {
            "table_type": "delta",
        },
    }
    with pytest.raises(ValueError, match="expected 'ICEBERG'"):
        GlueIcebergTable.from_table_info(glue_catalog, table_info)


###
# schema tests
###


def test_convert_glue_schema():
    columns = [
        {"Name": "bool_col", "Type": "boolean"},
        {"Name": "byte_col", "Type": "byte"},
        {"Name": "short_col", "Type": "short"},
        {"Name": "int_col", "Type": "integer"},
        {"Name": "long_col", "Type": "long"},
        {"Name": "bigint_col", "Type": "bigint"},
        {"Name": "float_col", "Type": "float"},
        {"Name": "double_col", "Type": "double"},
        {"Name": "decimal_col", "Type": "decimal"},
        {"Name": "string_col", "Type": "string"},
        {"Name": "timestamp_col", "Type": "timestamp"},
        {"Name": "date_col", "Type": "date"},
    ]

    # Test schema conversion
    assert _convert_glue_schema(columns) == Schema.from_pydict(
        {
            "bool_col": DataType.bool(),
            "byte_col": DataType.int8(),
            "short_col": DataType.int16(),
            "int_col": DataType.int32(),
            "long_col": DataType.int64(),
            "bigint_col": DataType.int64(),
            "float_col": DataType.float32(),
            "double_col": DataType.float64(),
            "decimal_col": DataType.decimal128(precision=38, scale=18),
            "string_col": DataType.string(),
            "timestamp_col": DataType.timestamp(timeunit="us", timezone="UTC"),
            "date_col": DataType.date(),
        }
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from daft.lazy_import import LazyImport

if TYPE_CHECKING:
    import fsspec
    import numpy as np
    import pandas as pd
    import PIL.Image as pil_image
    import pyarrow as pa
    import pyarrow.csv as pacsv
    import pyarrow.dataset as pads
    import pyarrow.flight as flight
    import pyarrow.fs as pafs
    import pyarrow.json as pajson
    import pyarrow.parquet as pq
else:
    fsspec = LazyImport("fsspec")
    np = LazyImport("numpy")
    pd = LazyImport("pandas")
    pil_image = LazyImport("PIL.Image")
    pa = LazyImport("pyarrow")
    pacsv = LazyImport("pyarrow.csv")
    pads = LazyImport("pyarrow.dataset")
    pafs = LazyImport("pyarrow.fs")
    pajson = LazyImport("pyarrow.json")
    pq = LazyImport("pyarrow.parquet")
    flight = LazyImport("pyarrow.flight")

unity_catalog = LazyImport("daft.unity_catalog")

__all__ = [
    "flight",
    "fsspec",
    "np",
    "pa",
    "pacsv",
    "pads",
    "pafs",
    "pajson",
    "pd",
    "pil_image",
    "pq",
    "unity_catalog",
]



from __future__ import annotations

import pytest

from daft.expressions import col, lit
from daft.recordbatch import MicroPartition

REGEX = r"^\d"


@pytest.mark.parametrize(
    ["expr", "data", "expected"],
    [
        (col("col").str.extract(REGEX), ["123", "456", "789", "abc"], ["1", "4", "7", None]),
        (col("col").str.extract(lit(REGEX)), ["123", "456", "789", "abc"], ["1", "4", "7", None]),
        (
            col("col").str.extract(col("emptystrings") + lit(REGEX)),
            ["123", "456", "789", "abc"],
            ["1", "4", "7", None],
        ),
    ],
)
def test_series_utf8_extract(expr, data, expected) -> None:
    table = MicroPartition.from_pydict({"col": data, "emptystrings": ["", "", "", ""]})
    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"col": expected}



from __future__ import annotations

import os
from contextlib import contextmanager

import pytest
import ray

import daft
from daft import udf
from daft.datatype import DataType
from daft.internal.gpu import cuda_visible_devices
from tests.conftest import get_tests_daft_runner_name

pytestmark = pytest.mark.skipif(
    get_tests_daft_runner_name() == "native", reason="Native runner does not support GPU tests yet"
)


@contextmanager
def reset_runner_with_gpus(num_gpus, monkeypatch):
    """If current runner does not have enough GPUs, create a new runner with mocked GPU resources."""
    if len(cuda_visible_devices()) < num_gpus:
        if get_tests_daft_runner_name() == "ray":
            try:
                ray.shutdown()
                ray.init(num_gpus=num_gpus)
                yield
            finally:
                ray.shutdown()
                ray.init()
        else:
            try:
                monkeypatch.setenv("CUDA_VISIBLE_DEVICES", ",".join(str(i) for i in range(num_gpus)))

                # Need to reset runner to recompute resources
                original_runner = daft.context.get_context()._runner
                daft.context.get_context()._runner = None
                yield
            finally:
                daft.context.get_context()._runner = original_runner
    else:
        yield


@pytest.mark.parametrize("concurrency", [1, 2])
@pytest.mark.parametrize("num_gpus", [1, 2])
def test_actor_pool_udf_cuda_env_var(monkeypatch, concurrency, num_gpus):
    with reset_runner_with_gpus(concurrency * num_gpus, monkeypatch):

        @udf(return_dtype=DataType.string(), num_gpus=num_gpus)
        class GetCudaVisibleDevices:
            def __init__(self):
                self.cuda_visible_devices = os.environ["CUDA_VISIBLE_DEVICES"]

            def __call__(self, data):
                assert os.environ["CUDA_VISIBLE_DEVICES"] == self.cuda_visible_devices

                import time

                time.sleep(0.1)

                return [self.cuda_visible_devices] * len(data)

        GetCudaVisibleDevices = GetCudaVisibleDevices.with_concurrency(concurrency)

        df = daft.from_pydict({"x": [1, 2, 3, 4]})
        df = df.repartition(4)
        df = df.select(GetCudaVisibleDevices(df["x"]))

        result = df.to_pydict()

        unique_visible_devices = set(result["x"])
        assert len(unique_visible_devices) == concurrency

        all_devices = (",".join(unique_visible_devices)).split(",")
        assert len(all_devices) == concurrency * num_gpus


def test_actor_pool_udf_fractional_gpu(monkeypatch):
    with reset_runner_with_gpus(1, monkeypatch):

        @udf(return_dtype=DataType.string(), num_gpus=0.5)
        class FractionalGpuUdf:
            def __init__(self):
                self.cuda_visible_devices = os.environ["CUDA_VISIBLE_DEVICES"]

            def __call__(self, data):
                assert os.environ["CUDA_VISIBLE_DEVICES"] == self.cuda_visible_devices

                import time

                time.sleep(0.1)

                return [self.cuda_visible_devices] * len(data)

        FractionalGpuUdf = FractionalGpuUdf.with_concurrency(2)

        df = daft.from_pydict({"x": [1, 2]})
        df = df.into_partitions(2)
        df = df.select(FractionalGpuUdf(df["x"]))

        result = df.to_pydict()

        unique_visible_devices = set(result["x"])
        assert len(unique_visible_devices) == 1



from __future__ import annotations

import random

import pandas as pd
import pytest

from daft import Window, col
from daft.functions import dense_rank, rank, row_number
from tests.conftest import assert_df_equals, get_tests_daft_runner_name

pytestmark = pytest.mark.skipif(
    get_tests_daft_runner_name() != "native", reason="Window tests only run on native runner"
)


def test_row_number_function(make_df):
    df = make_df(
        {"category": ["A", "A", "A", "B", "B", "B", "C", "C"], "sales": [100, 200, 50, 500, 100, 300, 250, 150]}
    )

    window_spec = Window().partition_by("category").order_by("sales", desc=False)

    result = df.select(
        col("category"), col("sales"), row_number().over(window_spec).alias("row_number_sales")
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "sales": [100, 200, 50, 500, 100, 300, 250, 150],
        "row_number_sales": [2, 3, 1, 3, 1, 2, 2, 1],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()), check_dtype=False)


def test_row_number_function_desc(make_df):
    df = make_df(
        {"category": ["A", "A", "A", "B", "B", "B", "C", "C"], "sales": [100, 200, 50, 500, 100, 300, 250, 150]}
    )

    window_spec = Window().partition_by("category").order_by("sales", desc=True)

    result = df.select(
        col("category"), col("sales"), row_number().over(window_spec).alias("row_number_sales")
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "sales": [100, 200, 50, 500, 100, 300, 250, 150],
        "row_number_sales": [2, 1, 3, 1, 3, 2, 1, 2],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()), check_dtype=False)


def test_multiple_window_partitions(make_df):
    """Test multiple window functions with different partition keys using random numbers.

    Creates a dataset with 900 rows (100 rows each for A1, A2, A3, B1, B2, B3, C1, C2, C3)
    and verifies row_number ordering within each partition.
    """
    import random

    random.seed(42)

    data = []
    all_numbers = list(range(1, 1001))
    random.shuffle(all_numbers)
    number_idx = 0

    for letter in ["A", "B", "C"]:
        for num in ["1", "2", "3"]:
            group_values = sorted(all_numbers[number_idx : number_idx + 3])
            number_idx += 3

            for value in group_values:
                data.append({"letter": letter, "num": num, "value": value})

    df = make_df(data)

    letter_window = Window().partition_by("letter").order_by("value", desc=False)
    num_window = Window().partition_by("num").order_by("value", desc=False)
    combined_window = Window().partition_by(["letter", "num"]).order_by("value", desc=False)

    result = df.select(
        col("letter"),
        col("num"),
        col("value"),
        row_number().over(letter_window).alias("letter_row_number"),
        row_number().over(num_window).alias("num_row_number"),
        row_number().over(combined_window).alias("combined_row_number"),
    ).collect()

    result_dict = result.to_pydict()

    for letter in ["A", "B", "C"]:
        letter_indices = [i for i, ltr in enumerate(result_dict["letter"]) if ltr == letter]
        letter_values = [result_dict["value"][i] for i in letter_indices]

        sorted_letter_values = sorted(letter_values)
        value_to_rank = {val: i + 1 for i, val in enumerate(sorted_letter_values)}

        for idx in letter_indices:
            value = result_dict["value"][idx]
            expected_rank = value_to_rank[value]
            actual_rank = result_dict["letter_row_number"][idx]
            assert (
                actual_rank == expected_rank
            ), f"Incorrect row number for letter {letter}, value {value}: got {actual_rank}, expected {expected_rank}"

    for num in ["1", "2", "3"]:
        num_indices = [i for i, n in enumerate(result_dict["num"]) if n == num]
        num_values = [result_dict["value"][i] for i in num_indices]

        sorted_num_values = sorted(num_values)
        value_to_rank = {val: i + 1 for i, val in enumerate(sorted_num_values)}
        for idx in num_indices:
            value = result_dict["value"][idx]
            expected_rank = value_to_rank[value]
            actual_rank = result_dict["num_row_number"][idx]
            assert (
                actual_rank == expected_rank
            ), f"Incorrect row number for num {num}, value {value}: got {actual_rank}, expected {expected_rank}"

    for letter in ["A", "B", "C"]:
        for num in ["1", "2", "3"]:
            combined_indices = [
                i
                for i, (ltr, n) in enumerate(zip(result_dict["letter"], result_dict["num"]))
                if ltr == letter and n == num
            ]
            combined_values = [result_dict["value"][i] for i in combined_indices]

            sorted_combined_values = sorted(combined_values)
            value_to_rank = {val: i + 1 for i, val in enumerate(sorted_combined_values)}
            for idx in combined_indices:
                value = result_dict["value"][idx]
                expected_rank = value_to_rank[value]
                actual_rank = result_dict["combined_row_number"][idx]
                assert (
                    actual_rank == expected_rank
                ), f"Incorrect row number for {letter}{num}, value {value}: got {actual_rank}, expected {expected_rank}"


def test_multi_window_agg_functions(make_df):
    """Test multiple window functions with different partition keys.

    Tests window functions with two different partition specifications:
    1. Partitioning by both category and group
    2. Partitioning by just category

    Using row_number ordering.
    """
    data = [
        {"category": "A", "group": 1, "value": 15},
        {"category": "A", "group": 1, "value": 25},
        {"category": "A", "group": 2, "value": 35},
        {"category": "A", "group": 2, "value": 45},
        {"category": "B", "group": 1, "value": 55},
        {"category": "B", "group": 1, "value": 65},
        {"category": "B", "group": 2, "value": 75},
        {"category": "B", "group": 2, "value": 85},
    ]

    df = make_df(data)

    multi_partition_window = Window().partition_by(["category", "group"]).order_by("value", desc=False)
    single_partition_window = Window().partition_by("category").order_by("value", desc=False)

    result = df.select(
        col("category"),
        col("group"),
        col("value"),
        row_number().over(multi_partition_window).alias("multi_row_number"),
        row_number().over(single_partition_window).alias("single_row_number"),
    ).collect()

    result_dict = result.to_pydict()

    for category in ["A", "B"]:
        for group in [1, 2]:
            indices = [
                i
                for i, (cat, grp) in enumerate(zip(result_dict["category"], result_dict["group"]))
                if cat == category and grp == group
            ]

            values = [result_dict["value"][i] for i in indices]
            sorted_values = sorted(values)
            value_to_rank = {val: i + 1 for i, val in enumerate(sorted_values)}

            for idx in indices:
                value = result_dict["value"][idx]
                expected_rank = value_to_rank[value]
                actual_rank = result_dict["multi_row_number"][idx]
                assert (
                    actual_rank == expected_rank
                ), f"Incorrect multi-partition row number for {category}/{group}, value {value}: got {actual_rank}, expected {expected_rank}"

    for category in ["A", "B"]:
        indices = [i for i, cat in enumerate(result_dict["category"]) if cat == category]

        values = [result_dict["value"][i] for i in indices]
        sorted_values = sorted(values)
        value_to_rank = {val: i + 1 for i, val in enumerate(sorted_values)}

        for idx in indices:
            value = result_dict["value"][idx]
            expected_rank = value_to_rank[value]
            actual_rank = result_dict["single_row_number"][idx]
            assert (
                actual_rank == expected_rank
            ), f"Incorrect single-partition row number for {category}, value {value}: got {actual_rank}, expected {expected_rank}"


def test_rank_function_single_row_per_group(make_df):
    """Test rank function with single row per group."""
    data = [
        {"category": "A", "value": 10},
        {"category": "B", "value": 20},
        {"category": "C", "value": 30},
    ]

    df = make_df(data)

    window_spec = Window().partition_by("category").order_by("value", desc=False)

    result = df.select(col("category"), col("value"), rank().over(window_spec).alias("rank")).collect()

    expected = {
        "category": ["A", "B", "C"],
        "value": [10, 20, 30],
        "rank": [1, 1, 1],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=["category"], check_dtype=False)

    result = df.select(col("category"), col("value"), dense_rank().over(window_spec).alias("rank")).collect()

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=["category"], check_dtype=False)


def test_rank_function_no_order_by(make_df):
    """Test rank function when order_by columns are the same as group_by."""
    data = [
        {"category": "A", "value": 10},
        {"category": "A", "value": 20},
        {"category": "A", "value": 30},
        {"category": "B", "value": 40},
        {"category": "B", "value": 50},
        {"category": "C", "value": 60},
    ]

    df = make_df(data)

    window_spec = Window().partition_by("category").order_by("category", desc=False)

    result = df.select(col("category"), col("value"), rank().over(window_spec).alias("rank")).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "C"],
        "value": [10, 20, 30, 40, 50, 60],
        "rank": [1, 1, 1, 1, 1, 1],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=["category", "value"], check_dtype=False)

    result = df.select(col("category"), col("value"), dense_rank().over(window_spec).alias("rank")).collect()

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=["category", "value"], check_dtype=False)


def test_multiple_rank_functions(make_df):
    """Test multiple rank functions over different window specifications.

    Creates a dataset with category (A, B, C) and subcategory (1, 2, 3) groups,
    each containing values with ties. Tests row_number, rank, and dense_rank
    functions over different window specifications.
    """
    import random

    random.seed(42)

    data = []

    for category in ["A", "B", "C"]:
        for subcategory in [1, 2, 3]:
            values = [random.randint(1, 10) for _ in range(8)]

            for value in values:
                data.append({"category": category, "subcategory": subcategory, "value": value})

    random.shuffle(data)
    df = make_df(data)

    category_window = Window().partition_by("category").order_by("value", desc=False)
    subcategory_window = Window().partition_by("subcategory").order_by("value", desc=False)
    combined_window = Window().partition_by(["category", "subcategory"]).order_by("value", desc=False)

    result = df.select(
        col("category"),
        col("subcategory"),
        col("value"),
        row_number().over(category_window).alias("row_number_by_category"),
        row_number().over(subcategory_window).alias("row_number_by_subcategory"),
        row_number().over(combined_window).alias("row_number_combined"),
        rank().over(category_window).alias("rank_by_category"),
        rank().over(subcategory_window).alias("rank_by_subcategory"),
        rank().over(combined_window).alias("rank_combined"),
        dense_rank().over(category_window).alias("dense_rank_by_category"),
        dense_rank().over(subcategory_window).alias("dense_rank_by_subcategory"),
        dense_rank().over(combined_window).alias("dense_rank_combined"),
    ).collect()

    result_df = result.to_pandas()

    def validate_rank_function(df, partition_cols, rank_col, rank_type):
        partition_groups = df.groupby(partition_cols)

        for partition_key, group in partition_groups:
            sorted_group = group.sort_values("value").reset_index(drop=True)

            value_to_ranks = {}
            for _, row in sorted_group.iterrows():
                value = row["value"]
                rank = row[rank_col]

                if value not in value_to_ranks:
                    value_to_ranks[value] = []
                value_to_ranks[value].append(rank)

            if rank_type == "row_number":
                all_ranks = sorted([r for ranks in value_to_ranks.values() for r in ranks])
                assert len(all_ranks) == len(
                    set(all_ranks)
                ), f"Row numbers should be unique within partition {partition_key}"

                assert all_ranks == list(range(1, len(all_ranks) + 1)), f"Row numbers should be sequential: {all_ranks}"

            elif rank_type == "rank":
                for value, ranks in value_to_ranks.items():
                    assert len(set(ranks)) == 1, f"All instances of value {value} should have the same rank: {ranks}"

                unique_value_ranks = {value: ranks[0] for value, ranks in value_to_ranks.items()}

                prev_value = None
                prev_rank = 0
                for value in sorted(unique_value_ranks.keys()):
                    rank = unique_value_ranks[value]
                    assert (
                        rank > prev_rank
                    ), f"Rank should increase for larger values: {value} has rank {rank}, previous rank was {prev_rank}"

                    if prev_value is not None:
                        expected_min_rank = prev_rank + len(value_to_ranks[prev_value])
                        assert (
                            rank >= expected_min_rank
                        ), f"Rank for value {value} should be at least {expected_min_rank}, got {rank}"

                    prev_value = value
                    prev_rank = rank

            elif rank_type == "dense_rank":
                for value, ranks in value_to_ranks.items():
                    assert (
                        len(set(ranks)) == 1
                    ), f"All instances of value {value} should have the same dense rank: {ranks}"

                unique_value_ranks = {value: ranks[0] for value, ranks in value_to_ranks.items()}

                unique_values = sorted(unique_value_ranks.keys())
                unique_ranks = [unique_value_ranks[v] for v in unique_values]

                assert unique_ranks[0] == 1, f"First dense rank should be 1, got {unique_ranks[0]}"

                for i in range(1, len(unique_ranks)):
                    assert (
                        unique_ranks[i] == unique_ranks[i - 1] + 1
                    ), f"Dense rank should increase by 1 for each distinct value: {unique_values[i]} has rank {unique_ranks[i]}, previous was {unique_ranks[i-1]}"

    for partition, col_prefix in [
        (["category"], "by_category"),
        (["subcategory"], "by_subcategory"),
        (["category", "subcategory"], "combined"),
    ]:
        validate_rank_function(result_df, partition, f"row_number_{col_prefix}", "row_number")
        validate_rank_function(result_df, partition, f"rank_{col_prefix}", "rank")
        validate_rank_function(result_df, partition, f"dense_rank_{col_prefix}", "dense_rank")


def test_multi_ordering_combinations(make_df):
    """Test row numbering with all possible ordering combinations."""
    random.seed(42)

    all_point_coordinates = []
    for x in range(10):
        for y in range(10):
            all_point_coordinates.append((x, y))

    test_data = []
    for group in ["A", "B", "C"]:
        selected_points = random.sample(all_point_coordinates, 10)
        for x, y in selected_points:
            test_data.append({"group": group, "x": x, "y": y})

    df = make_df(test_data)

    window_spec1 = Window().partition_by("group").order_by(["x", "y"], desc=[False, False])  # x asc, y asc
    window_spec2 = Window().partition_by("group").order_by(["x", "y"], desc=[False, True])  # x asc, y desc
    window_spec3 = Window().partition_by("group").order_by(["x", "y"], desc=[True, False])  # x desc, y asc
    window_spec4 = Window().partition_by("group").order_by(["x", "y"], desc=[True, True])  # x desc, y desc
    window_spec5 = Window().partition_by("group").order_by(["y", "x"], desc=[False, False])  # y asc, x asc
    window_spec6 = Window().partition_by("group").order_by(["y", "x"], desc=[False, True])  # y asc, x desc
    window_spec7 = Window().partition_by("group").order_by(["y", "x"], desc=[True, False])  # y desc, x asc
    window_spec8 = Window().partition_by("group").order_by(["y", "x"], desc=[True, True])  # y desc, x desc

    result = df.select(
        col("group"),
        col("x"),
        col("y"),
        row_number().over(window_spec1).alias("row_number_1"),
        row_number().over(window_spec2).alias("row_number_2"),
        row_number().over(window_spec3).alias("row_number_3"),
        row_number().over(window_spec4).alias("row_number_4"),
        row_number().over(window_spec5).alias("row_number_5"),
        row_number().over(window_spec6).alias("row_number_6"),
        row_number().over(window_spec7).alias("row_number_7"),
        row_number().over(window_spec8).alias("row_number_8"),
    ).collect()

    result_dict = result.to_pydict()

    row_number_mapping = {
        (0, 1, 1): 1,  # x asc, y asc
        (0, 1, -1): 2,  # x asc, y desc
        (0, -1, 1): 3,  # x desc, y asc
        (0, -1, -1): 4,  # x desc, y desc
        (1, 1, 1): 5,  # y asc, x asc
        (1, 1, -1): 6,  # y asc, x desc
        (1, -1, 1): 7,  # y desc, x asc
        (1, -1, -1): 8,  # y desc, x desc
    }

    for group in ["A", "B", "C"]:
        for primary_key in [0, 1]:  # 0 = x, 1 = y
            for primary_desc in [1, -1]:  # 1 = asc, -1 = desc
                for secondary_desc in [1, -1]:  # 1 = asc, -1 = desc
                    row_number_index = row_number_mapping[(primary_key, primary_desc, secondary_desc)]

                    points_with_row_numbers = [
                        (result_dict["x"][i], result_dict["y"][i], result_dict[f"row_number_{row_number_index}"][i])
                        for i, grp in enumerate(result_dict["group"])
                        if grp == group
                    ]

                    sorted_points = sorted(
                        points_with_row_numbers,
                        key=lambda p: (p[primary_key] * primary_desc, p[not primary_key] * secondary_desc),
                    )

                    for i, (x, y, actual_row_num) in enumerate(sorted_points):
                        expected_row_num = i + 1
                        assert expected_row_num == actual_row_num, (
                            f"Incorrect row number for group {group}, "
                            f"primary key {['x', 'y'][primary_key]} {('asc' if primary_desc == 1 else 'desc')}, "
                            f"secondary key {['x', 'y'][not primary_key]} {('asc' if secondary_desc == 1 else 'desc')}: "
                            f"expected {expected_row_num}, got {actual_row_num}"
                        )



from __future__ import annotations

from typing import Any

from daft.daft import PyWindowBoundary as _PyWindowBoundary
from daft.daft import WindowFrame as _WindowFrame
from daft.daft import WindowSpec as _WindowSpec
from daft.expressions import Expression
from daft.utils import ManyColumnsInputType, column_inputs_to_expressions


class Window:
    """Describes how to partition data and in what order to apply the window function.

    This class provides a way to specify window definitions for window functions.
    Window functions operate on a group of rows (called a window frame) and return
    a result for each row based on the values in its window frame.

    Examples:
        >>> from daft import Window, col
        >>>
        >>> # Basic window aggregation with a single partition column:
        >>> window_spec = Window().partition_by("category")
        >>> df = df.select(
        ...     col("value").sum().over(window_spec).alias("category_total"),
        ...     col("value").mean().over(window_spec).alias("category_avg"),
        ... )
        >>>
        >>> # Partitioning by multiple columns:
        >>> window_spec = Window().partition_by(["department", "category"])
        >>> df = df.select(col("sales").sum().over(window_spec).alias("dept_category_total"))
        >>>
        >>> # Using window aggregations in expressions:
        >>> window_spec = Window().partition_by("category")
        >>> df = df.select((col("value") / col("value").sum().over(window_spec)).alias("pct_of_category"))
    """

    # Class-level constants for frame boundaries
    unbounded_preceding = _PyWindowBoundary.unbounded_preceding()
    unbounded_following = _PyWindowBoundary.unbounded_following()
    current_row = _PyWindowBoundary.offset(0)

    def __init__(self) -> None:
        self._spec = _WindowSpec.new()

    def partition_by(self, *cols: ManyColumnsInputType) -> Window:
        """Partitions the dataset by one or more columns or expressions.

        Args:
            *cols: Columns or expressions on which to partition data.
                   Can be column names as strings, Expression objects, or iterables of these.

        Returns:
            Window: A window specification with the given partitioning.

        Raises:
            ValueError: If no partition columns are specified.

        Examples:
            >>> from daft import Window, col
            >>> # Partition by a single column 'category'
            >>> window_spec = Window().partition_by("category")
            >>> # Partition by multiple columns 'department' and 'region'
            >>> window_spec_multi = Window().partition_by("department", "region")
        """
        if not cols:
            raise ValueError("At least one partition column must be specified")

        expressions = []
        for c in cols:
            expressions.extend(column_inputs_to_expressions(c))

        if not expressions:
            raise ValueError("At least one partition column must be specified")

        window = Window()
        window._spec = self._spec.with_partition_by([expr._expr for expr in expressions])
        return window

    def order_by(self, *cols: ManyColumnsInputType, desc: bool | list[bool] = False) -> Window:
        """Orders rows within each partition by specified columns or expressions.

        Args:
            *cols: Columns or expressions to determine ordering within the partition.
                   Can be column names as strings, Expression objects, or iterables of these.
            desc: Sort descending (True) or ascending (False). Can be a single boolean value applied to all columns,
                 or a list of boolean values corresponding to each column. Default is False (ascending).

        Returns:
            Window: A window specification with the given ordering.

        Examples:
            >>> from daft import Window, col
            >>> # Order by 'date' ascending (default)
            >>> window_spec = Window().partition_by("category").order_by("date")
            >>> # Order by 'sales' descending
            >>> window_spec_desc = Window().partition_by("category").order_by("sales", desc=True)
            >>> # Order by 'date' ascending and 'sales' descending
            >>> window_spec_multi = Window().partition_by("category").order_by("date", "sales", desc=[False, True])
        """
        expressions = []
        for c in cols:
            expressions.extend(column_inputs_to_expressions(c))

        if isinstance(desc, bool):
            desc_flags = [desc] * len(expressions)
        else:
            if len(desc) != len(expressions):
                raise ValueError("Length of descending flags must match number of order by columns")
            desc_flags = desc

        window = Window()
        window._spec = self._spec.with_order_by([expr._expr for expr in expressions], desc_flags)
        return window

    def rows_between(
        self,
        start: int | _PyWindowBoundary,
        end: int | _PyWindowBoundary,
        min_periods: int = 1,
    ) -> Window:
        """Restricts each window to a row-based frame between start and end boundaries.

        This defines a sliding window based on row offsets relative to the current row.

        Args:
            start: Boundary definitions for the start of the window. Can be:

                *   ``Window.unbounded_preceding``: Include all rows before the current row.
                *   ``Window.current_row``: Start at the current row.
                *   Integer value (e.g., ``-3``): A negative integer indicates the number of rows *preceding* the current row.
                *   Integer value (e.g., ``1``): A positive integer indicates the number of rows *following* the current row.
            end: Boundary definitions for the end of the window. Can be:

                *   ``Window.unbounded_following``: Include all rows after the current row.
                *   ``Window.current_row``: End at the current row.
                *   Integer value (e.g., ``1``): A positive integer indicates the number of rows *following* the current row.
                *   Integer value (e.g., ``-1``): A negative integer indicates the number of rows *preceding* the current row.
            min_periods: Minimum number of rows required in the window frame to compute a result (default = 1).
                If fewer rows exist in the frame, the function returns NULL.

        Returns:
            Window: A window specification with the given row-based frame bounds.

        Examples:
            >>> from daft import Window
            >>> # Frame includes the current row and the 2 preceding rows
            >>> window_spec = Window().partition_by("cat").order_by("val").rows_between(-2, Window.current_row)
            >>> # Frame includes all rows from the beginning of the partition up to the current row
            >>> cum_window = (
            ...     Window()
            ...     .partition_by("cat")
            ...     .order_by("val")
            ...     .rows_between(Window.unbounded_preceding, Window.current_row)
            ... )
            >>> # Frame includes the preceding row, current row, and following row
            >>> sliding_window = Window().partition_by("cat").order_by("val").rows_between(-1, 1)
            >>> # Frame includes the current row and the 3 following rows, requiring at least 2 rows
            >>> lookahead_window = (
            ...     Window().partition_by("cat").order_by("val").rows_between(Window.current_row, 3, min_periods=2)
            ... )
        """
        if isinstance(start, int):
            start = _PyWindowBoundary.offset(start)
        if isinstance(end, int):
            end = _PyWindowBoundary.offset(end)

        frame = _WindowFrame(
            start=start,
            end=end,
        )

        new_window = Window()
        new_window._spec = self._spec.with_frame(frame).with_min_periods(min_periods)
        return new_window

    def range_between(
        self,
        start: Any,
        end: Any,
        min_periods: int = 1,
    ) -> Window:
        """Restricts each window to a range-based frame between start and end boundaries.

        This defines a window frame based on a range of values relative to the current row's
        value in the ordering column. Requires exactly one `order_by` column, which must be
        numeric or temporal type.

        Args:
            start: Boundary definition for the start of the window's range. Can be:

                *   ``Window.unbounded_preceding``: Include all rows with order value <= current row's order value + start offset.
                *   ``Window.current_row``: Start range at the current row's order value.
                *   Offset value (e.g., ``-10``, ``datetime.timedelta(days=-1)``): The start of the range is defined as
                    `current_row_order_value + start`. The type of the offset must match the order-by column type.
                    Negative values indicate a lower bound *less than* the current row's value. Positive values indicate a lower bound
                    *more than* the current row's value.
            end: Boundary definition for the end of the window's range. Syntax is similar to `start`.
                Negative values indicate an upper bound *less than* the current row's value. Positive values indicate an upper bound
                *more than* the current row's value.
            min_periods: Minimum number of rows required in the window frame to compute a result (default = 1).
                If fewer rows exist in the frame, the function returns NULL.

        Returns:
            Window: A window specification with the given range-based frame bounds.

        Raises:
            NotImplementedError: This feature is not yet implemented.

        Examples:
            >>> from daft import Window, col
            >>> import datetime
            >>> # Assume df has columns 'sensor_id', 'timestamp', 'reading'
            >>> # Frame includes rows within 10 units *before* the current row's reading
            >>> val_window = Window().partition_by("sensor_id").order_by("reading").range_between(-10, Window.current_row)
            >>> # Frame includes rows from 1 day before to 1 day after the current row's timestamp
            >>> time_window = (
            ...     Window()
            ...     .partition_by("sensor_id")
            ...     .order_by("timestamp")
            ...     .range_between(datetime.timedelta(days=-1), datetime.timedelta(days=1))
            ... )
        """
        if isinstance(start, _PyWindowBoundary):
            start_boundary = start
        else:
            start_expr = Expression._to_expression(start)
            start_boundary = _PyWindowBoundary.range_offset(start_expr._expr)

        if isinstance(end, _PyWindowBoundary):
            end_boundary = end
        else:
            end_expr = Expression._to_expression(end)
            end_boundary = _PyWindowBoundary.range_offset(end_expr._expr)

        frame = _WindowFrame(
            start=start_boundary,
            end=end_boundary,
        )

        new_window = Window()
        new_window._spec = self._spec.with_frame(frame).with_min_periods(min_periods)
        return new_window



"""
Taken from: https://github.com/cloudpipe/cloudpickle/blob/master/cloudpickle/compat.py
"""

from __future__ import annotations

import pickle  # noqa: F401

# Pickler will the C implementation in CPython and the Python
# implementation in PyPy
from pickle import Pickler  # noqa: F401



from __future__ import annotations

import pytest

from daft.datatype import DataType as dt
from daft.datatype import DataTypeLike


@pytest.mark.parametrize(
    "text,dtype,expected",
    [
        # json null -> null
        ("null", dt.null(), None),
        # json null -> T
        ("null", dt.bool(), None),
        ("null", dt.int64(), None),
        ("null", dt.string(), None),
        # json number -> bool
        ("false", dt.bool(), False),
        ("true", dt.bool(), True),
        # json number -> integer
        ("42", dt.int8(), 42),
        ("42", dt.int16(), 42),
        ("42", dt.int32(), 42),
        ("42", dt.uint8(), 42),
        ("42", dt.uint16(), 42),
        ("42", dt.uint32(), 42),
        ("42", dt.uint64(), 42),
        # json number -> float (can't compare floats accurately)
        # ("3.14", dt.float64(), 3.14),
        # ("3.14", dt.float32(), 3.14),
        # json number -> decimal ()
        # ("3.14", dt.decimal128(precision=10, scale=2), 3.14),
        # text -> json string
        ('"hello"', dt.string(), "hello"),
    ],
)
def test_deserialize_json_with_scalars(deserialize, text, dtype, expected):
    assert deserialize([text], "json", dtype)[0] == expected


@pytest.mark.parametrize(
    "data_type_like",
    [
        "STRUCT<name STRING, age INT64>",
        dt.struct({"name": dt.string(), "age": dt.int64()}),
    ],
)
def test_deserialize_json_with_struct(deserialize, data_type_like: DataTypeLike):
    items = [
        '{"name": "Alice", "age": 30}',
        '{"name": "Bob", "age": 25}',
        '{"name": "Charlie", "age": 35}',
    ]
    assert deserialize(items, "json", data_type_like) == [
        {"name": "Alice", "age": 30},
        {"name": "Bob", "age": 25},
        {"name": "Charlie", "age": 35},
    ]


@pytest.mark.parametrize(
    "data_type_like",
    [
        "INT64[]",
        dt.list(dt.int64()),
    ],
)
def test_deserialize_json_with_list(deserialize, data_type_like: DataTypeLike):
    items = [
        "[1, 2, 3]",
        "[4, 5, 6]",
    ]
    assert deserialize(items, "json", data_type_like) == [
        [1, 2, 3],
        [4, 5, 6],
    ]


@pytest.mark.skip("our arrow2 does not yet support json deserialize on the map type.")
def test_deserialize_json_with_map(deserialize):
    items = [
        '{"a": 1, "b": 2}',
        '{"c": 3, "d": 4}',
    ]
    # MAP<STRING, INT64>
    dtype = dt.map(dt.string(), dt.int64())
    assert deserialize(items, "json", dtype) == [
        {"a": 1, "b": 2},
        {"c": 3, "d": 4},
    ]


def test_try_deserialize_json(deserialize, try_deserialize):
    items = [
        "1",
        "true",
        "null",
        '"abc"',
        "3.14",
    ]
    assert try_deserialize(items, "json", dt.int64()) == [1, 1, None, None, 3]
    assert try_deserialize(items, "json", dt.bool()) == [None, True, None, None, None]
    assert try_deserialize(items, "json", dt.string()) == ["1", "true", None, "abc", "3.14"]


def test_deserialize_json_with_missing_fields(deserialize):
    items = [
        '{"name": "Alice", "age": 30}',
        '{ "age": 25}',
        '{"name": "Charlie"}',
    ]
    dtype = dt.struct({"name": dt.string(), "age": dt.int64()})
    assert deserialize(items, "json", dtype) == [
        {"name": "Alice", "age": 30},
        {"name": None, "age": 25},
        {"name": "Charlie", "age": None},
    ]



from __future__ import annotations

import pytest

from daft.expressions import col, lit
from daft.recordbatch import MicroPartition

REGEX = r"\d+"


@pytest.mark.parametrize(
    ["expr", "data", "expected"],
    [
        (col("col").str.extract_all(REGEX), ["1 2 3", "4 5 6", "a b c"], [["1", "2", "3"], ["4", "5", "6"], []]),
        (col("col").str.extract_all(lit(REGEX)), ["1 2 3", "4 5 6", "a b c"], [["1", "2", "3"], ["4", "5", "6"], []]),
        (
            col("col").str.extract_all(col("emptystrings") + lit(REGEX)),
            ["1 2 3", "4 5 6", "a b c"],
            [["1", "2", "3"], ["4", "5", "6"], []],
        ),
    ],
)
def test_series_utf8_extract_all(expr, data, expected) -> None:
    table = MicroPartition.from_pydict({"col": data, "emptystrings": ["", "", ""]})
    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"col": expected}



from __future__ import annotations

import pytest

import daft


@pytest.fixture()
def df():
    return daft.from_pydict(
        {
            "text": ["g1", "g1", "g2", "g3", "g3", "g1"],
            "n": [1, 2, 3, 3, 4, 100],
        }
    )


def test_orderby_basic(df):
    df = daft.sql("""
        SELECT * from df order by n
    """)

    assert df.collect().to_pydict() == {
        "text": ["g1", "g1", "g2", "g3", "g3", "g1"],
        "n": [1, 2, 3, 3, 4, 100],
    }


def test_orderby_compound(df):
    df = daft.sql("""
        SELECT * from df order by n, text
    """)

    assert df.collect().to_pydict() == {
        "text": ["g1", "g1", "g2", "g3", "g3", "g1"],
        "n": [1, 2, 3, 3, 4, 100],
    }


def test_orderby_desc(df):
    df = daft.sql("""
        SELECT n from df order by n desc
    """)

    assert df.collect().to_pydict() == {
        "n": [100, 4, 3, 3, 2, 1],
    }


def test_orderby_groupby(df):
    df = daft.sql("""
        SELECT
            text,
            count(*) as count_star
        from df
        group by text
        order by count_star DESC
    """)

    assert df.collect().to_pydict() == {
        "text": ["g1", "g3", "g2"],
        "count_star": [3, 2, 1],
    }


def test_orderby_groupby_expr(df):
    df = daft.sql("""
SELECT
    text,
    count(*) as count_star
from df
group by text
order by count(*) DESC
    """)

    assert df.collect().to_pydict() == {"text": ["g1", "g3", "g2"], "count_star": [3, 2, 1]}


def test_groupby_orderby_non_final_expr(df):
    df = daft.sql("""
        SELECT
            text,
            count(*) as count_star
        from df
        group by text
        order by sum(n) ASC
    """)

    assert df.collect().to_pydict() == {
        "text": ["g2", "g3", "g1"],
        "count_star": [1, 2, 3],
    }


def test_groupby_orderby_count_star(df):
    df = daft.sql("""
        SELECT
            text,
            sum(n) as n
        from df
        group by text
        order by count(*) ASC
    """)

    assert df.collect().to_pydict() == {"text": ["g2", "g3", "g1"], "n": [3, 7, 103]}



from __future__ import annotations

import pytest

import daft


@pytest.mark.integration()
def test_url_download_public_azure(azure_storage_public_config) -> None:
    data = {"urls": ["az://public-anonymous/mvp.parquet"]}
    df = daft.from_pydict(data)
    df = df.with_column("data", df["urls"].url.download(io_config=azure_storage_public_config))

    data = df.to_pydict()
    assert len(data["data"]) == 1
    for databytes in data["data"]:
        assert databytes is not None
